"""
Helper functions for CLI commands.

Author: Parker Hicks
Date: 2025-11-21

Last updated: 2025-11-21 by Parker Hicks
"""

import os
from pathlib import Path
import json

def set_verbosity(quiet: bool):
    """Return the opposite of quiet."""
    if quiet:
        return False
    return True

def dir_to_nested_dict(path):
    """Get the rough directory sturcture of database, edit output for _validate.py"""
    result = {}
    
    try:
        for item in os.listdir(path):
            item_path = os.path.join(path, item)
            
            if os.path.isdir(item_path):
                result[item] = dir_to_nested_dict(item_path)
            else:
                if '__files__' not in result:
                    result['__files__'] = []
                result['__files__'].append(item)
    except PermissionError:
        result['__error__'] = 'Permission denied'
    
    return print(json.dumps(result, indent=2))



"""
CLI command to search the ontology database for terms. Relies
on metahq_core.search.search() to do the actual searching.

Author: Faisal Alquaddoomi
Date: 2025-09-25

Last updated: 2025-11-24 by Parker Hicks
"""

import click
from metahq_core.search import NoResultsFound
from metahq_core.search import search as core_search
from metahq_core.util.progress import get_console
from metahq_core.util.supported import get_log_dir, get_ontology_search_db

from metahq_cli.logger import setup_logger
from metahq_cli.util.common_args import logging_args
from metahq_cli.util.helpers import set_verbosity

DEFAULT_TOP_HITS = 3


@click.command(name="search", context_settings={"help_option_names": ["-h", "--help"]})
@click.option("--query", type=str, required=True, help="Search query")
@click.option(
    "--db",
    "-b",
    type=str,
    default="default",
    help="DuckDB file",
    hidden=True,
)
@click.option(
    "--type",
    "-c",
    type=click.Choice(["disease", "tissue", "celltype", "none"]),
    default=None,
    help="Filter by type",
)
@click.option(
    "--ontology",
    "-o",
    type=click.Choice(["MONDO", "UBERON", "CL"]),
    default=None,
    help="Filter by ontology",
)
@click.option(
    "--max-results",
    "-k",
    type=int,
    default=DEFAULT_TOP_HITS,
    help="Number of results to show",
)
@click.option(
    "--scores", "-s", is_flag=True, default=False, help="Include scores in results"
)
@click.option(
    "--extended",
    "-e",
    is_flag=True,
    default=False,
    help="Include ontology and type in results",
)
@click.option(
    "--scopes", "-x", is_flag=True, default=False, help="Include scopes in synonym list"
)
@logging_args
def search(
    query,
    db,
    type,
    ontology,
    max_results,
    scores,
    extended,
    scopes,
    log_level,
    quiet,
):
    """Search for terms in the ontology database."""
    verbose = set_verbosity(quiet)
    logger = setup_logger(
        __name__, console=get_console(), level=log_level, log_dir=get_log_dir()
    )

    if db == "default":
        db = get_ontology_search_db()

    try:
        results = core_search(
            query,
            db=db,
            type=type,
            ontology=ontology,
            k=max_results,
            logger=logger,
            verbose=verbose,
        )

        # format results like so:
        # 1. MONDO:0001234
        #    name: Some Disease
        #    synonyms: synonym1 | synonym2 | synonym3

        for idx, result in enumerate(results.iter_rows(named=True)):
            print(f"{idx+1}. {result['term_id']}")
            print(f"    name: {result['name']}")

            # if extended=True, include ontology and type
            if extended:
                print(f"    ontology: {result['ontology']}")
                print(f"    type: {result['type']}")

            # if scores=True, include score
            if scores:
                print(f"    score: {result['score']:.4f}")

            # if scopes=True, include synonyms' scopes in the synonym list
            # otherwise, just return the pipe-delimited list of synonyms
            if result["synonyms"]:
                syns = " | ".join(
                    [
                        f"{s[0]} (scope: {s[1]})" if scopes else s[0]
                        for s in result["synonyms"]
                    ]
                )
                print(f"    synonyms: {syns}")
            # add newline at end of output
            print("")

    except NoResultsFound as e:
        print(e)



"""
Logger setup.

Author: Parker Hicks
Date: 2025-10-16

Last updated: 2026-02-04 by Parker Hicks
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from rich.logging import RichHandler

from metahq_cli.util.checkers import check_loglevel

if TYPE_CHECKING:
    from rich.console import Console


class ColoredFormatter(logging.Formatter):
    """Console color logger formatter."""

    COLORS = {
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "bold red",
    }

    def format(self, record):
        levelname = record.levelname
        color = self.COLORS.get(levelname, "white")
        record.levelname = f"[{color}][{levelname}][/{color}]"
        return super().format(record)


def setup_logger(
    name: str,
    console: Console,
    log_dir: str | Path,
    level: int | str = logging.INFO,
) -> logging.Logger:
    """Sets up a logger.

    Arguments:
        name (str):
            Logger name.

        console (rich.Console):
            Console object to send logging information to.

        log_dir (str | Path):
            Path to logging directory.

        level (int):
            Logging level.

    Returns:
        Configured logger.

    """
    logger = logging.getLogger(name)
    _level = check_loglevel(level)

    if logger.hasHandlers():
        return logger

    logger.setLevel(_level)

    # rich console handler
    console_handler = RichHandler(
        console=console,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        markup=True,
        show_level=False,
        show_time=True,
        show_path=False,
        log_time_format="[%x %X]",
        omit_repeated_times=False,
    )

    console_formatter = ColoredFormatter(
        fmt="{levelname} {message}",
        style="{",
    )
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    # file handler
    file_handler = logging.FileHandler(
        Path(log_dir) / "log.log",
        encoding="utf-8",
    )
    # file formatter
    file_formatter = logging.Formatter(
        fmt="[{asctime}] [{levelname}] {message}",
        style="{",
        datefmt="%Y-%m-%d %H:%M",
    )
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    return logger



from metahq_cli.commands.delete import delete
from metahq_cli.commands.retrieve import retrieve_commands
from metahq_cli.commands.search import search
from metahq_cli.commands.setup import setup
from metahq_cli.commands.supported import supported
from metahq_cli.commands.validate import validate

__all__ = ["retrieve_commands", "setup", "search", "supported", "validate", "delete"]



"""
Templates for CLI errors and warnings.

Author: Parker Hicks
Date: 2025-09-25

Last updated: 2025-09-25
"""

import sys

import click


def warning(message):
    click.secho(f"WARNING: {message}", fg="yellow")


def error(message):
    click.secho(f"ERROR: {message}", fg="red")
    sys.exit(1)


class TruncatedList:
    """Truncate long lists for a less egregious display."""

    def __init__(self, data, max_show=4):
        self.data: list[str] = data
        self.max_show: int = max_show

    def __repr__(self):
        per_side = self.max_show // 2

        start = ", ".join(map(repr, self.data[:per_side]))
        end = ", ".join(map(repr, self.data[-per_side:]))

        return f"[{start}, ... , {end}] (n={len(self.data)})"



"""
Facilitates argument and curation parsing for metaHQ retrieval commands.

Author: Parker Hicks
Date: 2025-09-25

Last updated: 2025-12-05 by Parker Hicks
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import polars as pl
from metahq_core.query import Query
from metahq_core.util.exceptions import NoResultsFound
from metahq_core.util.supported import supported

from metahq_cli.util.messages import TruncatedList

if TYPE_CHECKING:
    import logging

    from metahq_core.curations.annotations import Annotations
    from metahq_core.curations.labels import Labels


@dataclass
class QueryConfig:
    """Storage for query parameters.

    Attributes:
        database (str):
            The name of a supported database within MetaHQ.

        attribute (str):
            A supported attribute within MetaHQ.

        level (str):
            A level of annotations (e.g., 'sample' or 'series').

        ecode (str):
            Evidence code (e.g., 'expert', 'crowd', 'any')

        species (str):
            A supported species within MetaHQ.

        tech (str):
            A supported technology within MetaHQ.
    """

    database: str
    attribute: str
    level: str
    ecode: str
    species: str
    tech: str


@dataclass
class CurationConfig:
    """Storage for curation parameters.

    Attributes:
        mode (str):
            A supported curation mode (e.g., 'annotate', 'label').

        terms (str):
            A list of terms to curate annotations for.

        ontology (str):
            An ontology to use for propagating annotations and assigning labels.
    """

    mode: str
    terms: list[str]
    ontology: str


@dataclass
class OutputConfig:
    """Storage for output parameters.

    Attributes:
        outfile (str | Path):
            Path to file to store annotations.

        fmt (Literal["json", "parquet", "csv", "tsv"]):
            Format of the output file.

        metadata (str):
            Comma-delimited string indicating which metadata fields to include.

        attribute (str):
            A supported attribute within MetaHQ.

        level (str):
            A level of annotations (e.g., 'sample' or 'series').
    """

    outfile: str | Path
    fmt: Literal["json", "parquet", "csv", "tsv"]
    metadata: str
    attribute: str
    level: str


class Retriever:
    """
    Queries, curates, and saves MetaHQ annotations for `metahq retrieve`.
    Exists to reduce redundancy in MetaHQ retrieve commands.

    Attributes:
        query_config: QueryConfig
            Parameters for querying.

        curation_config: CurationConfig
            Parameters for curating annotations.

        output_config: OutputConfig
            Parameters for saving curations.
    """

    def __init__(
        self, query_config, curation_config, output_config, logger, verbose=True
    ):
        self.query_config: QueryConfig = query_config
        self.curation_config: CurationConfig = curation_config
        self.output_config: OutputConfig = output_config

        self.log: logging.Logger = logger
        self.verbose: bool = verbose

        if verbose:
            self.log.debug(
                "Using configs:\n%s\n%s\n%s",
                self.query_config,
                self.curation_config,
                self.output_config,
            )

    def curate(self, annotations: Annotations) -> Annotations:
        """Mutate curations by specified mode.

        Arguments:
            annotations: Annotations
                A populated Annotations object.

        Returns:
            A populated Annotations or Labels object given the specified curation mode.

        Raises:
            NoResultsFound: If there are no annotations for a set of query parameters.
        """
        if annotations.n_indices == 0:
            msg = "No annotations for any terms. Try using different conditions."
            self.log.error(msg)
            raise NoResultsFound(msg)

        if self.verbose:
            self.log.info("Curating...")

        return self._curate_by_mode(annotations)

    def query(self):
        """Performs a MetaHQ query"""
        if self.verbose:
            return self._query_verbose()

        return self._query_silent()

    def retrieve(self):
        """Performs the retrieval pipeline: query -> curate -> save."""
        curation = self.query()
        curation = self.curate(curation)
        self.save_curation(curation)

    def save_curation(self, curation: Annotations | Labels):
        """Saves the curation.

        Arguments:
            curation (Annotations | Labels):
                A populated Annotations or Labels object to save.
        """
        self._save(curation)

    def _curate_by_mode(self, curation: Annotations) -> Annotations | Labels:
        """Apply the appropriate curation method to queried annotations."""
        if self.curation_config.mode == "direct":
            return self._direct_annotations(curation)

        if self.curation_config.mode == "annotate":
            return self._propagate_annotations(curation, mode=0)

        if self.curation_config.mode == "label":
            return self._propagate_annotations(curation, mode=1)

        msg = (
            "Expected mode in %s, got %s.",
            supported("modes"),
            self.curation_config.mode,
        )

        if self.verbose:
            self.log.error(msg)

        raise ValueError(msg)

    def _direct_annotations(self, curation: Annotations) -> Annotations:
        """Identify and return terms in the query that have annotations."""

        available_terms = self._filter_missing_entities(curation)

        return curation.select(available_terms).filter(
            pl.any_horizontal(pl.col(available_terms) == 1)
        )

    def _propagate_annotations(
        self, curation: Annotations, mode: Literal[0, 1]
    ) -> Annotations | Labels:
        """Wrapper for Annotations propagation."""

        result = curation.propagate(
            self.curation_config.terms,
            self.curation_config.ontology,
            mode=mode,
        )

        return result

    def _query(self) -> Annotations:
        if self.verbose:
            self.log.info("Querying...")

        return Query(
            database=self.query_config.database,
            attribute=self.query_config.attribute,
            level=self.query_config.level,
            ecode=self.query_config.ecode,
            species=self.query_config.species,
            technology=self.query_config.tech,
            logger=self.log,
            verbose=self.verbose,
        ).annotations()

    def _query_silent(self):
        return self._query()

    def _filter_missing_entities(self, curation: Annotations | Labels) -> list[str]:
        terms_with_anno = [
            term for term in self.curation_config.terms if term in curation.entities
        ]
        not_in_anno = [
            term for term in self.curation_config.terms if not term in terms_with_anno
        ]

        if len(not_in_anno) == len(self.curation_config.terms):
            msg = "No annotations for any terms. Try using different conditions."
            self.log.error(msg)
            raise NoResultsFound(msg)

        if self.verbose:
            if len(terms_with_anno) != len(self.curation_config.terms):
                if len(not_in_anno) > 10:
                    not_in_anno = TruncatedList(not_in_anno)
                self.log.warning(
                    "Queries: %s have no annotations. Try using different conditions.",
                    not_in_anno,
                )
        return terms_with_anno

    def _query_verbose(self):
        return self._query()

    def _save(self, curation):

        if self.verbose:
            self.log.info("Saving to %s...", self.output_config.outfile)

        curation.save(
            outfile=self.output_config.outfile,
            fmt=self.output_config.fmt,
            metadata=self.output_config.metadata,
            attribute=self.output_config.attribute,
            level=self.output_config.level,
        )



"""
Helper info for validate command.
"""

from pathlib import Path
import hashlib

def get_files_to_check(doi):
    if doi == "18462463":
        FILES_TO_CHECK = [
            [Path("annotations") / "combined__level-sample.bson", "07c761c25eeb37787b13e3f0be7ea4db"],
            [Path("annotations") / "combined__level-series.bson", "4a70f6c1a4cdd693cf33b50c561e960b"],
            [Path("metadata") / "metadata__level-sample.parquet", "2527ffbc7b7bacd6054841b395975b8b"],
            [Path("metadata") / "metadata__level-series.parquet", "0215ea97cf1e09e98a43d219b53b254f"],
            [Path("metadata") / "technologies.parquet", "a7cd45dc7db09d30fe35676d8d449b32"],
            [Path("ontology") / "ontology_search.duckdb", "5108270059472abc4a18e0ee3a7c68f2"],
            [Path("ontology") / "mondo" / "id_map.parquet", "37230bd391ec3af4be4f710ab2d76707"],
            [Path("ontology") / "mondo" / "names_synonyms.json", "24318fa528ded695943b53787c0c7b00"],
            [Path("ontology") / "mondo" / "relations.parquet", "9d460ad0eaa06d85767717e2f6456163"],
            [Path("ontology") / "mondo" / "systems.txt", "dbf3e5566b4dd80e458b2cd5813ad693"],
            [Path("ontology") / "uberon_ext" / "id_map.parquet", "a449bed0812bacfa35d9b83628da594e"],
            [Path("ontology") / "uberon_ext" / "names_synonyms.json", "9e9fd448715929351e8c0e4b9561d56c"],
            [Path("ontology") / "uberon_ext" / "relations.parquet", "a9c3cead75ac43c2be3b961d69d25d53"],
            [Path("ontology") / "uberon_ext" / "systems.txt", "8e2b4b0943ae52720463257d67ec8fbf"],  
        ]
    if doi == "17666183":
        FILES_TO_CHECK = [
            [Path("annotations") / "combined__level-sample.bson", "5d41627b6194b34e19bda2edf9289b6d"],
            [Path("annotations") / "combined__level-series.bson", "c4280f1432a001ebe031d6eb4d7b5c5c"],
            [Path("metadata") / "metadata__level-sample.parquet", "5ab5771210d31f6cc3d81c29fec20a7e"],
            [Path("metadata") / "metadata__level-series.parquet", "06629fd37a62905e79e31e8eaddf021d"],
            [Path("metadata") / "technologies.parquet", "a7cd45dc7db09d30fe35676d8d449b32"],
            [Path("ontology") / "ontology_search.duckdb", "5108270059472abc4a18e0ee3a7c68f2"],
            [Path("ontology") / "mondo" / "id_map.parquet", "37230bd391ec3af4be4f710ab2d76707"],
            [Path("ontology") / "mondo" / "names_synonyms.json", "24318fa528ded695943b53787c0c7b00"],
            [Path("ontology") / "mondo" / "relations.parquet", "9d460ad0eaa06d85767717e2f6456163"],
            [Path("ontology") / "mondo" / "systems.txt", "dbf3e5566b4dd80e458b2cd5813ad693"],
            [Path("ontology") / "uberon_ext" / "id_map.parquet", "a449bed0812bacfa35d9b83628da594e"],
            [Path("ontology") / "uberon_ext" / "names_synonyms.json", "9e9fd448715929351e8c0e4b9561d56c"],
            [Path("ontology") / "uberon_ext" / "relations.parquet", "a9c3cead75ac43c2be3b961d69d25d53"],
            [Path("ontology") / "uberon_ext" / "systems.txt", "b5327f3e6591768bcdb24b1b083cd0ea"],  
        ]
    return FILES_TO_CHECK

def md5_file(filepath):
    """Calculate MD5 checksum of a file"""
    hash_md5 = hashlib.md5()
    
    with open(filepath, 'rb') as f:
        # Read in chunks to handle large files efficiently
        for chunk in iter(lambda: f.read(4096), b''):
            hash_md5.update(chunk)
    
    return hash_md5.hexdigest()
    
def check_md5_match(config_doi, config_data_dir):
    """Check MD5 checksum match"""
    FILES_TO_CHECK =  get_files_to_check(config_doi)
    changed_files = []
    for afile in FILES_TO_CHECK:
        filepath = Path(config_data_dir) / afile[0]
        new_checksum =  md5_file(filepath)
        old_checksum = afile[1]
        if new_checksum != old_checksum:
            changed_files.append(filepath)
    return changed_files


"""
Downloader for the MetaHQ database stored on Zenodo.
Implemented in the `metahq setup` CLI command.

Author: Parker Hicks
Date: 2025-11-20

Last updated: 2026-02-05 by Parker Hicks
"""

import shutil
import sys
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import click
import requests
from metahq_core.util.io import checkdir
from metahq_core.util.progress import get_console, progress_bar
from metahq_core.util.supported import get_default_log_dir

from metahq_cli.logger import setup_logger
from metahq_cli.util.supported import (
    default_data_dir,
    metahq_dois,
    zenodo_files_dir,
    zenodo_records_url,
)

if TYPE_CHECKING:
    import logging


DEFAULT_OUTDIR: Path = default_data_dir()


@dataclass
class FileConfig:
    """Storage for MetaHQ database record information.

    Attributes:
    doi: str
        A valid MetaHQ Zenodo DOI.

    filename: str
        The name of the MetaHQ database .tar.gz file in
        the files for the Zenodo DOI.

    url: str
        The full URL to the MetaHQ database file in Zenodo.

    version: str
        The version of the MetaHQ database.

    outdir: Path
        Path to the output directory in the user's system.

    filesize: int | None:
        Size of the databse file in bytes.
    """

    doi: str
    filename: str
    url: str
    version: str
    outdir: Path
    filesize: int | None = None

    def make_outdir(self):
        """Check that outdir exists."""
        checkdir(self.outdir)

    @property
    def filename_stemmed(self) -> str:
        """Return file name without any extensions."""
        return str(self.filename).split(".", maxsplit=1)[0]

    @property
    def outfile(self) -> Path:
        """Return path to the database file on the user's system."""
        return self.outdir / self.filename

    @property
    def size_mb(self) -> float:
        """Return the filesize in MB."""
        if self.filesize is None:
            raise AttributeError("filesize not set.")

        return round(self.filesize / (1024**2), 2)


class Downloader:
    """Downloader for the MetaHQ database stored on Zenodo.

    Attributes:
        doi (str):
            Zenodo DOI for the MetaHQ databse.

        outdir (str):
            Path to the output directory in the user's system.

        logger (logging.Logger):
            Logger for process transparency.

        verbose (bool):
            Indicates if logs should be passed to stdout.

    Examples:
        >>> from metahq_cli.downloader import Downloader
        >>> downloader = Downloader('17663087')
        >>> downloader.get()
        >>> downloader.extract()
    """

    def __init__(
        self,
        doi,
        outdir: str | Path = DEFAULT_OUTDIR,
        logger=None,
        loglevel=20,
        logdir=get_default_log_dir(),
        verbose=True,
    ):
        self.config: FileConfig = self._make_config(doi, outdir)

        if logger is None:
            logger = setup_logger(
                __name__, level=loglevel, log_dir=logdir, console=get_console()
            )
        self.logger: logging.Logger = logger
        self.verbose: bool = verbose

        self._use_progress: bool = True

    def check_outdir_exists(self):
        """Check if the data directory exists."""
        if self.config.outdir.exists():
            if click.confirm(
                f"Data directory: {self.config.outdir} exists. Overwrite?",
                default=False,
            ):
                self.logger.info("Removing existing data directory...")
                shutil.rmtree(self.config.outdir)
                self.config.make_outdir()
            else:
                self.logger.info("Keeping existing data directory.")
                sys.exit("Terminating...")
        else:
            self.config.make_outdir()

    def extract(self):
        """Extract the tar archive."""
        if self.verbose:
            self.logger.info("Decompressing...")

        self._extract()

        tar_dir = self.config.outdir / self.config.filename_stemmed
        self._move_tar_contents(base_dir=self.config.outdir, tar_dir=tar_dir)

        if self.verbose:
            self.logger.info("Saved the MetaHQ database to %s.", self.config.outdir)

    def get(self, num_retries=10):
        """Downloads the database .tar.gz file from Zenodo."""
        self.check_outdir_exists()

        num_tries = 0
        one_more_try = True
        while num_tries <= num_retries - 2:
            num_tries += 1
            self.logger.info(f"On download attempt {num_tries} out of {num_retries}")
            try:
                self.get_stats()
                self._download()
                one_more_try = False
                break
            except:
                continue

        # try one more time in needed  to find the last error and report it
        if one_more_try:
            self.logger.info(f"On download attempt {num_retries} out of {num_retries}")
            try:
                self.get_stats()
                self._download()

            except requests.exceptions.ConnectionError:
                self._raise_connection_error()

            except requests.exceptions.Timeout:
                self._raise_timeout_error()

            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    self._raise_404_error()
                elif e.response.status_code == 403:
                    self._raise_403_error()
                else:
                    self._raise_http_error(e)

            except PermissionError:
                self._raise_permissions_error()

            except KeyboardInterrupt:
                self._raise_keyboard_interrupt()

            except Exception as e:
                self._raise_general_exception(e)

    def get_stats(self):
        """Check if file exists and retrieve file stats."""
        if self.verbose:
            self.logger.debug("Valid DOI: %s", self.config.doi)
            self.logger.debug("Downloading from URL: %s", self.config.url)
            self.logger.info("Checking file availability...")

        head = requests.head(self.config.url, allow_redirects=True, timeout=10)
        self.config.filesize = int(head.headers.get("content-length", 0))

        if self.verbose:
            self.logger.debug("Request status: %s", head.status_code)
            self.logger.debug("Request headers: %s", dict(head.headers))
            self.logger.debug("Final URL: %s", head.url)
            self.logger.debug("File size: %s MB", self.config.size_mb)

        if self.config.filesize == 0:
            self._use_progress = False

            if self.verbose:
                self.logger.warning("Could not determine file size.")
                self.logger.info("Downloading without progress bar.")

    @property
    def database_version(self) -> str:
        """Return the MetaHQ database version."""
        return self.config.version

    # ========================================
    # ======  downloaders
    # ========================================

    def _download(self):
        if self.verbose:
            self.logger.debug("Show progress bar: %s", self._use_progress)

        if self._use_progress:
            self._download_with_progress()
        else:
            self._download_no_progress()

    def _download_with_progress(self):
        response = requests.get(
            self.config.url,
            stream=True,
            allow_redirects=True,
            timeout=30,
            headers={
                "User-Agent": "<meta-hq>/v1 (https://github.com/krishnanlab/meta-hq)"
            },
        )
        with progress_bar(padding="    ") as progress:
            task = progress.add_task("Downloading", total=self.config.filesize)

            with open(self.config.outfile, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
                    progress.update(task, advance=len(chunk))

    def _download_no_progress(self):
        response = requests.get(
            self.config.url,
            stream=True,
            allow_redirects=True,
            timeout=30,
            headers={
                "User-Agent": "<meta-hq>/v1 (https://github.com/krishnanlab/meta-hq)"
            },
        )
        with open(self.config.outfile, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

    # ========================================
    # ======  tar extractors
    # ========================================

    def _extract(self):
        with tarfile.open(self.config.outfile, mode="r:gz") as tar:
            tar.extractall(
                path=self.config.outdir, members=tar.getmembers(), filter="data"
            )

    def _move_tar_contents(self, base_dir: Path, tar_dir: Path):
        target_dir = base_dir

        if self.verbose:
            self.logger.debug("Moving contents from %s to %s", tar_dir, target_dir)

        for item in tar_dir.iterdir():
            shutil.move(str(item), str(target_dir / item.name))

        if self.verbose:
            self.logger.debug("Deleting %s", tar_dir)

        shutil.rmtree(tar_dir)
        (base_dir / self.config.filename).unlink()

    # ========================================
    # ======  helpers
    # ========================================

    def _make_config(self, doi: str, outdir: str | Path) -> FileConfig:
        info: dict[str, str] = metahq_dois(doi)

        # looks like https://zenodo.org/records/<doi>/files/<filename>
        url: str = self._make_url(doi, info["filename"])

        return FileConfig(
            doi=doi,
            filename=info["filename"],
            outdir=Path(outdir),
            url=url,
            version=info["version"],
        )

    def _make_url(self, doi: str, filename: str) -> str:
        base_url = zenodo_records_url()
        files_dir = zenodo_files_dir()

        return f"{base_url}/{doi}/{files_dir}/{filename}/content"

    # ========================================
    # ======  error messages
    # ========================================

    def _raise_connection_error(self):
        self.logger.error(
            "Could not connect to Zenodo. Check your internet connection.",
        )
        sys.exit(1)

    def _raise_general_exception(self, e):
        self.logger.error("Unexpected error occurred: %s", e)
        sys.exit(1)

    def _raise_http_error(self, e):
        self.logger.error("HTTP %s: %s", e.response.status_code, e)
        sys.exit(1)

    def _raise_keyboard_interrupt(self):
        self.logger.error("Download cancelled by user.")
        if self.config.outfile.exists():
            self.config.outfile.unlink()
            self.logger.info("Removed partial download: %s", self.config.outfile)
        sys.exit(130)

    def _raise_permissions_error(self):
        self.logger.error(
            "Permission denied writing to '%s'. Check file permissions.",
            self.config.outfile,
        )
        sys.exit(1)

    def _raise_timeout_error(self):
        self.logger.error(
            "Download timed out. The server may be slow or unreachable.",
        )
        sys.exit(1)

    def _raise_403_error(self):
        self.logger.error(
            "Access forbidden (403). The file may be restricted.",
        )
        sys.exit(1)

    def _raise_404_error(self):
        self.logger.error(
            "File not found (404). Check that the URL is correct (hint: use --loglevel debug).",
        )
        sys.exit(1)



"""
Settings for supported CLI options.

Author: Parker Hicks
Date: 2025-09-05

Last updated: 2026-02-02 by Parker Hicks
"""

from pathlib import Path

LATEST_DATABASE: dict[str, str] = {"version": "v1.0.1", "doi": "18462463"}
DOIS = {
    "17663087": {"version": "v1.0.0-alpha", "filename": "metahq.tar.gz"},
    "17666183": {"version": "v1.0.0", "filename": "metahq_data.tar.gz"},
    "18462463": {"version": "v1.0.1", "filename": "metahq_data.tar.gz"},
}


def log_map() -> dict[str, int]:
    """Return mapping between log levels as text and their corresponding int values."""
    return {
        "notset": 0,
        "debug": 10,
        "info": 20,
        "warning": 30,
        "error": 40,
        "critical": 50,
    }


def required_filters() -> list[str]:
    return ["species", "tech", "ecode"]


def sample_ids() -> list[str]:
    return ["sample", "srr", "srx"]


def zenodo_records_url() -> str:
    """Return the url to Zenodo records."""
    return "https://zenodo.org/api/records"


def zenodo_files_dir() -> str:
    """Return the directory name for Zenodo files for a particular record."""
    return "files"


def metahq_dois(doi: str) -> dict[str, str]:
    """
    Return version and filename information for a MetaHQ database Zenodo DOI.

    Parameters
    ----------
    doi: str
        A valid Zenodo DOI for a MetaHQ database version.

    Returns
    -------
    Database version and the filename for the particular DOI.

    """
    if not doi in DOIS:
        available = list(DOIS.keys())
        raise ValueError(f"Expected doi in {available}, got {doi}.")

    return DOIS[doi]


def metahq_mother_dir() -> Path:
    """Return the name of the MetaHQ directory for every user.

    This directory will store the package config file and logs.
    """
    return Path.home() / "metahq"


def default_data_dir() -> Path:
    """Return the default name of the data directory."""
    return Path.home() / (".metahq_data")



"""
Class to take retrieval arguments, check them, and build the retrieval query.

Author: Parker Hicks
Date: 2025-10-16

Last updated: 2025-12-05 by Parker Hicks
"""

from typing import TYPE_CHECKING

import polars as pl
from metahq_core.util.exceptions import NoResultsFound
from metahq_core.util.supported import get_ontology_families

from metahq_cli.retriever import CurationConfig, OutputConfig, QueryConfig
from metahq_cli.util.checkers import (
    check_filter,
    check_filter_keys,
    check_format,
    check_if_txt,
    check_metadata,
    check_mode,
    check_outfile,
)
from metahq_cli.util.messages import TruncatedList
from metahq_cli.util.supported import required_filters

if TYPE_CHECKING:
    import logging


class Builder:
    """Class to build query, curation, and output configurations for `metahq retrieve`.
    Exists to support modularity and reduce redundnacy in the retrieval commands.
    """

    def __init__(self, logger, verbose=True):
        self.log: logging.Logger = logger
        self.verbose: bool = verbose

    def get_filters(self, filters: str) -> dict[str, str]:
        """Parses and checks requested filters.

        Arguments:
            filters (str):
                A comma-delimited string of supported MetaHQ filters.

        Returns:
            A dictionary of filter key, values.

        Examples:
            >>> from metahq_cli.retrieval_builder import Builder
            >>> builder = Builder()
            >>> filters = 'species=human,ecode=expert,tech=rnaseq'
            >>> builder.get_filters()
            {'species': 'human', 'ecode': 'expert', 'tech': 'rnaseq'}
        """
        _filters = self._parse_filters(filters)
        self.report_bad_filters(_filters)
        return _filters

    def parse_onto_terms(self, terms: list[str], reference: str) -> list[str]:
        """Collects passed query terms and checks if they are appropriate.

        Attributes:
            terms (list[str]):
                A list of ontology term IDs. Can be ontology IDs for tissues or diseases.
                (e.g., `['UBERON:0000948', 'UBERON:0000955']`).

            reference (str):
                An indication of what kind of terms these are (e.g., 'UBERON' or 'MONDO').

        Returns:
            Any terms that are supported within MetaHQ.

        Raises:
            NoResultsFound: If none of the terms are in the MetaHQ database.
        """
        available = (
            pl.scan_parquet(get_ontology_families(reference)["relations"])
            .collect_schema()
            .names()
        )

        if terms == "all":
            return available

        parsed = self._parse(terms, available)

        # check if terms are missing
        diff = abs(len(terms) - len(parsed))
        if diff > 0:
            missing = list(set(terms) - set(parsed))
            if len(missing) > 10:
                missing = TruncatedList(missing)

            # fail if all are missing
            if diff == len(terms):
                msg = (
                    f"""Terms: {missing} have no annotations for: {reference.upper()}"""
                )

                if self.verbose:
                    self.log.error(msg)

                raise NoResultsFound(msg)

            # only raise warning if some are missing
            if self.verbose:
                self.log.warning(
                    "No annotations for input terms: %s",
                    missing,
                )

        return parsed

    def query_config(
        self, db: str, attribute: str, level: str, filters: dict[str, str]
    ) -> QueryConfig:
        """Construct a query configuration.

        Query parameters are checked in the metahq_core.query module.

        Arguments:
            db (str):
                The name of a supported database within MetaHQ.

            attribute (str):
                A supported attribute within MetaHQ.

            level (str):
                A level of annotations (e.g., `'sample'` or `'series'`).

            filters (dict[str, str]):
                Filters parsed by `Builder.get_filters`.

        Returns:
            A populated `QueryConfig`.
        """
        check_filter("ecodes", filters["ecode"])
        check_filter("species", filters["species"])
        check_filter("technologies", filters["tech"])

        return QueryConfig(
            database=db,
            attribute=attribute,
            level=level,
            ecode=filters["ecode"],
            species=filters["species"],
            tech=filters["tech"],
        )

    def curation_config(self, terms: str, mode: str, ontology: str) -> CurationConfig:
        """Construct a curation configuration.

        Attributes:
            terms (str):
                A list of terms to curate annotations for.

            mode (str):
                A supported curation mode (e.g., `'annotate'`, `'label'`).

            ontology (str):
                An ontology to use for propagating annotations and assigning labels.

        Returns:
            A populated `CurationConfig`.
        """
        if ontology == "sex":
            return self.make_sex_curation(terms, mode)

        if ontology == "age":
            return self.make_age_curation(terms, mode)

        if terms == "all":
            _terms = self.parse_onto_terms(terms, ontology)
        else:
            _terms = check_if_txt(terms)

        if isinstance(_terms, str):
            _terms = _terms.split(",")

        _terms = self.parse_onto_terms(_terms, ontology)

        return CurationConfig(mode, _terms, ontology)

    def output_config(
        self,
        outfile: str,
        fmt: str,
        metadata: str,
        level: str,
        attribute: str,
    ) -> OutputConfig:
        """Construct an output configuration.

        Attributes:
            outfile (str | Path):
                Path to file to store annotations.

            fmt (Literal["json", "parquet", "csv", "tsv"]):
                Format of the output file.

            metadata (str):
                Comma-delimited string indicating which metadata fields to include.

            level (str):
                Annotation level to check if any of the requested metadata fields are
                    available for the requested level.

        Returns:
            A populated `OutputConfig`.
        """
        check_metadata(level, metadata)
        check_format(fmt)
        check_outfile(outfile)

        return OutputConfig(outfile, fmt, metadata, attribute, level)

    def make_age_curation(self, terms: str, mode: str) -> CurationConfig:
        """Makes an age-specific CurationConfig."""
        _terms = check_if_txt(terms)
        check_mode("age", mode)

        if _terms == "all":
            from metahq_core.util.supported import supported

            _terms = supported("age_groups")

        elif isinstance(_terms, str):
            _terms = _terms.split(",")

        else:
            self.log.error("Invalid input: %s", terms)

        return CurationConfig(mode, _terms, ontology="age")

    def make_sex_curation(self, terms: str, mode: str):
        """Sex-specific curation."""
        _terms = check_if_txt(terms)
        check_mode("sex", mode)

        if _terms == "all":
            from metahq_core.util.supported import sexes

            _terms = sexes()

        elif isinstance(_terms, str):
            _terms = _terms.split(",")

        else:
            self.log.error("Invalid input: %s", terms)

        _terms = self._map_sex_to_id(_terms)
        return CurationConfig(mode, _terms, ontology="sex")

    def _map_sex_to_id(self, terms: list[str]):
        """Map male to M and female to F if passed."""
        opt = {"male": "M", "female": "F"}

        result = []
        for term in terms:
            if term in ["male", "female"]:
                result.append(opt[term])
            else:
                result.append(term)

        return result

    def _parse(self, terms: list[str], available: list[str]) -> list[str]:
        return [term for term in terms if term in available]

    def _parse_filters(self, filters: str) -> dict[str, str]:
        as_list: list[list[str]] = [f.split("=") for f in filters.split(",")]
        as_dict: dict[str, str] = {f[0]: f[1] for f in as_list}

        not_in_filters: list[str] = []
        for key in required_filters():
            if key not in as_dict:
                not_in_filters.append(key)

        if len(not_in_filters) > 0:
            msg: str = f"Missing required filters {not_in_filters}."
            self.log.error(msg)
            raise RuntimeError(msg)

        return as_dict

    def report_bad_filters(self, filters: dict[str, str]):
        """Check filters and return improper filter parameters."""
        bad_filters = check_filter_keys(filters)
        if len(bad_filters) > 0:
            msg = ("Expected filters in %s, got %s.", required_filters(), bad_filters)

            if self.verbose:
                self.log.error(msg)

            raise ValueError(msg)



"""
CLI command to validate data hasen't changed.
"""

import click
from metahq_core.util.progress import console
from pathlib import Path
import hashlib

from metahq_cli.logger import setup_logger
from metahq_cli.util.common_args import logging_args
from metahq_core.util.supported import get_default_log_dir
from metahq_cli.util.helpers import set_verbosity

from metahq_core.util.supported import get_config, get_config_file
from metahq_cli.util._validate import check_md5_match

@click.command(name="validate", context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-l",
    "--log-dir",
    type=click.Path(),
    default="default",
    help="Path to directory storing logs. Default is `/home/path/MetaHQ`.",
)
@logging_args
def validate(log_level: str, log_dir: str, quiet: bool):
    """Validate files are unchanged"""
    if log_dir == "default":
        log_dir = str(get_default_log_dir())

    logger = setup_logger(__name__, level=log_level, log_dir=log_dir, console=console)
    verbose = set_verbosity(quiet)
    
    # read in config
    config = get_config()
    config_path = get_config_file()
    logger.info(f"The path to config file is found at: {config_path}")
    logger.info(f"Validating the files for the config options: {config}")
    
    changed_files = check_md5_match(config["zenodo_doi"], config["data_dir"])
    
    if len(changed_files) == 0:
        logger.info("Validation Passed: All files validated to be correct")
    elif len(changed_files) > 0:
        logger.warning(
            f"Validation Failed: The following files are not the same as when downloaded. "
            "Run `metahq setup` again."
        )
        for afile in changed_files:
            logger.warning(afile)


"""
Command to setup the metahq configuration.

Author: Parker Hicks
Date: 2025-09-05

Last updated: 2025-12-15 by Parker Hicks
"""

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import yaml
from metahq_core.util.progress import console
from metahq_core.util.supported import get_config_file_no_check

from metahq_cli.logger import setup_logger

if TYPE_CHECKING:
    import logging


CONFIG_FILE: Path = get_config_file_no_check()


class Config:
    """Class to store and assess and create the meta-hq configuration file.

    Attributes:
        version (str):
            A version name of the MetaHQ database (e.g., `'v0.1.0'`).

        zenodo_doi (str):
            A DOI of the MetaHQ database in Zenodo.

        data_dir (str):
            Path to where the MetaHQ data are stored.

        logs (str):
            Path to where the MetaHQ logs are stored.

        ok_keys (list[str]):
            Acceptable keys in the config.
    """

    def __init__(
        self,
        version,
        zenodo_doi,
        data_dir,
        logs,
        logger=None,
        loglevel=20,
        verbose=True,
    ):
        self.version: str = version
        self.zenodo_doi: str = zenodo_doi
        self.data_dir: str = data_dir
        self.logs: str = logs

        if logger is None:
            logger = setup_logger(
                __name__, level=loglevel, log_dir=logs, console=console
            )
        self.logger: logging.Logger = logger
        self.verbose = verbose

        self.ok_keys: list[str] = ["version", "zenodo_doi", "data_dir", "logs"]

    def check(self):
        """Checks if the meta-hq config exists. Initializes if not."""
        if not CONFIG_FILE.exists():
            if self.verbose:
                self.logger.debug("Initializing new config file %s", CONFIG_FILE)
            CONFIG_FILE.touch()

        else:
            if self.verbose:
                self.logger.debug("Config file exists: %s", CONFIG_FILE)
                self.logger.debug("Running config check...")
            # check existing
            if not self.is_acceptable_config():
                if self.verbose:
                    self.logger.warning(
                        "Incorrect configuration detected. Resetting with defaults."
                    )
                self.set_default()

    def is_acceptable_config(self):
        """Checks if config has correct structure."""
        config = self.load_config()

        if self.verbose:
            self.logger.debug("Existing config: %s", config)

        if config is None:
            if self.verbose:
                self.logger.debug("Existing config is empty.")
            return False

        return sorted(list(config.keys())) == sorted(self.ok_keys)

    def load_config(self) -> dict[str, str]:
        """Loads the meta-hq config file."""
        with open(CONFIG_FILE, "r", encoding="utf-8") as stream:
            try:
                return yaml.safe_load(stream)
            except yaml.YAMLError as e:
                sys.exit(str(e))

    def load_config_str(self) -> str:
        """Loads the meta-hq config file."""
        with open(CONFIG_FILE, "r", encoding="utf-8") as f:
            return f.read()

    def make_config(self) -> dict[str, str]:
        """Creates the config dictionary"""
        return {
            "version": self.version,
            "zenodo_doi": self.zenodo_doi,
            "data_dir": self.data_dir,
            "logs": self.logs,
        }

    def save_config(self, config: dict[str, str]):
        """Saves a config file.

        Arguments:
            config (dict[str, str]):
                A config with acceptable keys.

        """
        self.logger.info("Saving MetaHQ config to %s", CONFIG_FILE)
        with open(CONFIG_FILE, "w", encoding="utf-8") as stream:
            try:
                yaml.safe_dump(config, stream)
            except yaml.YAMLError as e:
                sys.exit(str(e))
        self.logger.info("Done!")

    def setup(self):
        """Main setup function."""
        self.check()
        new = self.initialize_config()
        self.save_config(new)

    def set_default(self):
        """Makes a default meta-hq config."""
        self.logger.info("Making config with default arguments.")
        self.make_config()

    def initialize_config(self):
        """Initialize the meta-hq config file."""
        config = {
            "version": self.version,
            "zenodo_doi": self.zenodo_doi,
            "data_dir": str(Path(self.data_dir).resolve()),
            "logs": self.logs,
        }

        if self.verbose:
            self.logger.debug("New configuration: %s", config)

        return config

    @property
    def path(self) -> str:
        """Returns `/path/to/config.yaml`"""
        return str(CONFIG_FILE)



"""
CLI entry point for meta-hq.

Author: Parker Hicks
Date: 2025-09-05

Last updated: 2025-09-05
"""

import click

from metahq_cli.commands import retrieve_commands, search, setup, supported, validate, delete


@click.group()
def main():
    pass


main.add_command(setup)
main.add_command(search)
main.add_command(supported)
main.add_command(retrieve_commands, name="retrieve")
main.add_command(validate)
main.add_command(delete)

if __name__ == "__main__":
    main()



"""
CLI command to retrieve annotations and labels from meta-hq.

Author: Parker Hicks
Date: 2025-09-05

Last updated: 2025-12-01 by Parker Hicks
"""

import click
from metahq_core.util.progress import get_console
from metahq_core.util.supported import get_log_dir, supported

from metahq_cli.logger import setup_logger
from metahq_cli.retrieval_builder import Builder
from metahq_cli.retriever import Retriever
from metahq_cli.util.common_args import (
    logging_args,
    ontology_retrieval_args,
    retrieval_args,
)
from metahq_cli.util.helpers import set_verbosity

AGE_GROUP_OPT = click.Choice(supported("age_groups") + ["all"])


def check_direct(mode: str, direct: bool, verbose: bool, logger) -> str:
    """Checks if direct flag was passed to return direct annotations.

    Parameters
    ----------
    mode: str
        The user's passed mode. Either annotate or label.

    direct: bool
        Value of the `direct` command argument.

    verbose: bool
        Verbosity yes or no.

    logger: logging.Logger
        Initialized Python Logger object.

    Returns
    -------
    'direct' if `direct` is True, otherwise mode.

    """
    if direct:
        if verbose:
            logger.info("Overriding passed mode '%s' with 'direct'.", mode)
        return "direct"
    return mode


# ===================================================
# ==== entry point
# ===================================================
@click.group
def retrieve_commands():
    """Retrieval commands for tissue, disease, sex, and age annotations."""


@retrieve_commands.command("age")
@click.option(
    "--terms",
    type=AGE_GROUP_OPT,
    default="all",
    help="Age groups to choose. Can combine like 'fetus,adult'.",
)
@retrieval_args
@logging_args
def retrieve_age(terms, level, fmt, metadata, filters, output, log_level, quiet):
    """Retrieval command for age group annotations."""
    if metadata == "default":
        metadata = level

    verbose = set_verbosity(quiet)
    log = setup_logger(
        __name__, console=get_console(), level=log_level, log_dir=get_log_dir()
    )

    builder = Builder(logger=log, verbose=verbose)

    # parse and check filters
    filters = builder.get_filters(filters)

    # make configs
    query_config = builder.query_config("geo", "age", level, filters)
    curation_config = builder.curation_config(terms, "direct", "age")
    output_config = builder.output_config(
        output, fmt, metadata, level=level, attribute="age"
    )

    # retrieve
    retriever = Retriever(
        query_config, curation_config, output_config, logger=log, verbose=verbose
    )
    retriever.retrieve()


@retrieve_commands.command("diseases")
@click.option("--terms", type=str, default="MONDO:0004994,MONDO:0018177")
@retrieval_args
@ontology_retrieval_args
@logging_args
def retrieve_diseases(
    terms, level, mode, fmt, metadata, filters, output, log_level, quiet, direct
):
    """Retrieval command for disease ontology terms."""
    if metadata == "default":
        metadata = level

    verbose = set_verbosity(quiet)
    log = setup_logger(
        __name__, console=get_console(), level=log_level, log_dir=get_log_dir()
    )

    # hidden from user. Used to test annotation quality.
    mode = check_direct(mode, direct, verbose, log)
    builder = Builder(logger=log, verbose=verbose)

    # parse and check filters
    filters = builder.get_filters(filters)

    # make configs
    query_config = builder.query_config("geo", "disease", level, filters)
    curation_config = builder.curation_config(terms, mode, "mondo")
    output_config = builder.output_config(
        output, fmt, metadata, level=level, attribute="disease"
    )

    # retrieve
    retriever = Retriever(
        query_config, curation_config, output_config, logger=log, verbose=verbose
    )
    retriever.retrieve()


@retrieve_commands.command("sex")
@click.option("--terms", type=str, default="male,female")
@retrieval_args
@logging_args
def retrieve_sex(terms, level, fmt, metadata, filters, output, log_level, quiet):
    """Retrieval command for sex annotations."""
    if metadata == "default":
        metadata = level

    verbose = set_verbosity(quiet)
    log = setup_logger(
        __name__, console=get_console(), level=log_level, log_dir=get_log_dir()
    )

    builder = Builder(logger=log, verbose=verbose)

    # parse and check filters
    filters = builder.get_filters(filters)

    # make configs
    query_config = builder.query_config("geo", "sex", level, filters)
    curation_config = builder.curation_config(terms, "direct", "sex")
    output_config = builder.output_config(
        output, fmt, metadata, level=level, attribute="sex"
    )

    # retrieve
    retriever = Retriever(
        query_config, curation_config, output_config, logger=log, verbose=verbose
    )
    retriever.retrieve()


@retrieve_commands.command("tissues")
@click.option("--terms", type=str, default="UBERON:0000948,UBERON:0000955")
@retrieval_args
@ontology_retrieval_args
@logging_args
def retrieve_tissues(
    terms, level, mode, fmt, metadata, filters, output, log_level, quiet, direct
):
    """Retrieval command for tissue ontology terms."""
    if metadata == "default":
        metadata = level

    verbose = set_verbosity(quiet)
    log = setup_logger(
        __name__, console=get_console(), level=log_level, log_dir=get_log_dir()
    )

    # hidden from user. Used to test annotation quality.
    mode = check_direct(mode, direct, verbose, log)
    builder = Builder(logger=log, verbose=verbose)

    # parse and check filters
    filters = builder.get_filters(filters)

    # make configs
    query_config = builder.query_config("geo", "tissue", level, filters)
    curation_config = builder.curation_config(terms, mode, "uberon")
    output_config = builder.output_config(
        output, fmt, metadata, level=level, attribute="tissue"
    )

    # retrieve
    retriever = Retriever(
        query_config, curation_config, output_config, logger=log, verbose=verbose
    )
    retriever.retrieve()



"""
Definition of common arguments across CLI commands. Exists to
remove redundancy and improve readability in CLI command
definitions.

Author: Parker Hicks
Date: 2025-11-24

Last updated: 2025-11-24 by Parker Hicks
"""

from functools import wraps

import click
from metahq_core.util.supported import supported

FMT_OPT = click.Choice(supported("formats"))
LEVEL_OPT = click.Choice(supported("levels"))
LOGLEVEL_OPT = click.Choice(supported("log_levels"))
MODE_OPT = click.Choice(supported("modes"))


def logging_args(command):
    """Decorator to assign logging arguments across CLI commands."""

    @click.option(
        "--log-level", type=LOGLEVEL_OPT, default="info", help="Logging level."
    )
    @click.option(
        "--quiet",
        is_flag=True,
        default=False,
        help="No log or console output if applied.",
    )
    @wraps(command)
    def wrapper(*args, **kwargs):
        return command(*args, **kwargs)

    return wrapper


def retrieval_args(command):
    """Decorator to assign common arguments across retrieval commands."""

    @click.option(
        "--level", type=LEVEL_OPT, default="sample", help="GEO annotation level."
    )
    @click.option(
        "--filters",
        type=str,
        default="species=human,ecode=expert,tech=rnaseq",
        help="Filters for species, ecode, and technology. Run `metahq supported` for options.",
    )
    @click.option(
        "--output",
        type=click.Path(),
        default="annotations.parquet",
        help="Path to outfile.",
    )
    @click.option("--fmt", type=FMT_OPT, default="parquet")
    @click.option("--metadata", type=str, default="default")
    @wraps(command)
    def wrapper(*args, **kwargs):
        return command(*args, **kwargs)

    return wrapper


def ontology_retrieval_args(command):
    """
    Decorator to assign common arguments specific to retrieval for
    ontology-based queries like tissues and diseases
    """

    @click.option(
        "--mode",
        type=MODE_OPT,
        default="annotate",
        help="Retrieve annotations or labels.",
    )
    @click.option(
        "--direct",
        is_flag=True,
        default=False,
        help="Get direct annotations. Hidden because not practical.",
        hidden=True,
    )
    @wraps(command)
    def wrapper(*args, **kwargs):
        return command(*args, **kwargs)

    return wrapper



"""
CLI command to delete old version of MetaHQ
"""

import click
import os
import shutil
import sys
from metahq_core.util.progress import console
from pathlib import Path

from metahq_cli.logger import setup_logger
from metahq_cli.util.common_args import logging_args
from metahq_core.util.supported import get_default_log_dir
from metahq_cli.util.helpers import set_verbosity

from metahq_core.util.supported import get_config, get_config_file

@click.command(name="delete", context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-l",
    "--log-dir",
    type=click.Path(),
    default="default",
    help="Path to directory storing logs. Default is `/home/path/MetaHQ`.",
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Will delete everything in MetaHQ directory",
)
@logging_args
def delete(log_level: str, log_dir: str, quiet: bool, all: bool):
    """Delete old version of MetaHQ"""
    if log_dir == "default":
        log_dir = str(get_default_log_dir())

    logger = setup_logger(__name__, level=log_level, log_dir=log_dir, console=console)
    verbose = set_verbosity(quiet)
    
    # read in config
    config = get_config()
    config_path = get_config_file()
    logger.info(f"The path to config file is found at: {config_path}")
    logger.info(f"Information in the config file is: {config}")
         
    if click.confirm(
        f"Delete MetaHQ database and config file or config directory?",
        default=False,
    ):
        logger.info("Deleting existing data directory...")
        shutil.rmtree(config["data_dir"])
        if all:
            logger.info("Deleting existing config directory...")
            shutil.rmtree(config_path.parent)
        else:
            logger.info("Deleting existing config file...")
            os.remove(config_path)
    else:
        logger.info("Keeping existing data directory and config file.")
        sys.exit("Terminating...")


"""
Checker functions for the MetaHQ CLI pipelines.

These checkers pull from two separate supported modules from
metahq_core and metahq_cli since there are different requirements
for CLI-based checks and MetaHQ core function checks.

Author: Parker Hicks
Date: 2025-09

Last updated: 2025-11-24 by Parker Hicks
"""

from pathlib import Path

from metahq_core.util.io import checkdir, load_txt
from metahq_core.util.supported import supported

from metahq_cli.util.messages import error
from metahq_cli.util.supported import log_map, required_filters


def check_filter_keys(filters: dict[str, str]):
    unaccaptable = []
    for f in filters:
        if f not in required_filters():
            unaccaptable.append(f)
    return unaccaptable


def check_binary(file: Path | str) -> bool:
    with open(file, "rb") as f:
        return b"\x00" in f.read(8000)


def check_if_txt(string: str) -> list[str] | str:
    if Path(string).is_file():
        if check_binary(Path(string)):
            error(f"Detected binary file: {string}. Please write terms to text file.")
        return load_txt(string)

    return string


def check_level(level: str):
    if level not in supported("levels"):
        error(f"Expected level in {supported('levels')}, got {level}.")


def check_loglevel(loglevel: int | str) -> int:
    """Check input log level."""
    mapping = log_map()
    if isinstance(loglevel, str) and (loglevel in mapping):
        return mapping[loglevel]
    if isinstance(loglevel, int) and (loglevel in mapping.values()):
        return loglevel

    raise ValueError(
        f"Invalid log level. Choose from {list(mapping.keys())} or {list(mapping.values())}."
    )


def check_metadata(level: str, metadata: str):
    _metadata = metadata.split(",")

    if level == "sample":
        _supported = supported("sample_metadata")

    elif level == "series":
        _supported = supported("series_metadata")

    else:
        check_level(level)
        exit(0)

    report_bad_entries("metadata", _supported, _metadata)


def check_format(fmt: str):
    _supported = supported("formats")
    if fmt not in _supported:
        error(f"Expected fmt argument in {_supported}, got {fmt}.")


def check_mode(task: str, mode: str):
    if (task in ["sex", "age"]) & (mode != "direct"):
        error(
            "Sex annotation queries must be direct annotations. Change to mode argument to 'direct'."
        )


def check_outfile(outfile: Path | str):
    _ = checkdir(outfile, is_file=True)


def check_filter(_filter: str, query: str):
    """Supports checking for ecode, technology, and species."""
    _supported = supported(_filter)
    if not query in _supported:
        error(f"Expected {_filter} in {_supported}, got {query}.")


def report_bad_entries(field: str, _supported: list[str], entries: list[str]):
    bad = []
    for entry in entries:
        if entry not in _supported:
            bad.append(entry)

    if len(bad) > 0:
        error(f"Bad arguments for {field}: {bad}. Expected arguments in {_supported}.")



"""
MetaHQ CLI command to display supported options for retrieval arguments.

Author: Parker Hicks
Date: 2025-09-29

Last updated: 2025-09-29 by Parker Hicks
"""

import click
from metahq_core.util.supported import _supported
from rich.console import Console
from rich.table import Table

console = Console()


@click.command
def supported():
    """Display all supported entities and their options."""
    table = Table(title="MetaHQ Supported Entities")
    table.add_column("Entity", style="cyan")
    table.add_column("Available", style="green")

    __supported = _supported()
    for i, (entity, avail) in enumerate(__supported.items()):
        avail = ", ".join(avail).strip()
        entity = " ".join(entity.split("_")).strip()
        table.add_row(entity, avail)
        if i < len(__supported) - 1:  # Don't add separator after last row
            table.add_section()

    console.print(table)



"""
Command to set up the meta-hq CLI.

Author: Parker Hicks
Date: 2025-09-05

Last updated: 2025-11-26 by Parker Hicks
"""

from pathlib import Path

import click
from metahq_core.util.progress import console
from metahq_core.util.supported import get_default_data_dir, get_default_log_dir

from metahq_cli.logger import setup_logger
from metahq_cli.setup.config import Config
from metahq_cli.setup.downloader import Downloader
from metahq_cli.util.common_args import logging_args
from metahq_cli.util.helpers import set_verbosity
from metahq_cli.util.supported import LATEST_DATABASE


@click.command(name="setup", context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-d",
    "--doi",
    type=str,
    default="latest",
    help="Zenodo DOI of the MetaHQ database. Default is `latest`.",
)
@click.option(
    "-o",
    "--data-dir",
    type=click.Path(),
    default="default",
    help="Path to directory to store the database. Default is `/home/path/.metahq_data`.",
)
@click.option(
    "-l",
    "--log-dir",
    type=click.Path(),
    default="default",
    help="Path to directory storing logs. Default is `/home/path/MetaHQ`.",
)
@logging_args
def setup(doi: str, data_dir: str, log_level: str, log_dir: str, quiet: bool):
    """Download the MetaHQ database and configure the CLI."""
    if log_dir == "default":
        log_dir = str(get_default_log_dir())

    logger = setup_logger(__name__, level=log_level, log_dir=log_dir, console=console)
    verbose = set_verbosity(quiet)

    if doi == "latest":
        doi = LATEST_DATABASE["doi"]

    if data_dir == "default":
        data_dir = str(get_default_data_dir())

    logger.info("Downloading MetaHQ database...")
    downloader = Downloader(
        doi, data_dir, logger=logger, logdir=str(log_dir), verbose=verbose
    )
    downloader.get()
    downloader.extract()

    logger.info("Configuring MetaHQ...")
    config = Config(
        downloader.database_version,
        doi,
        data_dir,
        log_dir,
        logger=logger,
        verbose=verbose,
    )
    config.setup()
